from ._fb_attrs import (  # NOQA: F401
    Flatbuffer as _Flatbuffer,
    FlatbufferEnum as _FlatbufferEnum,
    from_package as _from_package,
    from_package_enum as _from_package_enum,
    UNION_CL as UNION_CL,
    model_from_bytes as model_from_bytes,
    model_to_bytes as model_to_bytes,
)

Flatbuffer = _Flatbuffer
FlatbufferEnum = _FlatbufferEnum
from_package = _from_package
from_package_enum = _from_package_enum



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import struct

from flatbuffers import packer
from flatbuffers.compat import import_numpy, NumpyRequiredForThisFeature

np = import_numpy()

# For reference, see:
# https://docs.python.org/2/library/ctypes.html#ctypes-fundamental-data-types-2

# These classes could be collections.namedtuple instances, but those are new
# in 2.6 and we want to work towards 2.5 compatability.

cdef class Flags:
    def __init__(self, unsigned short bytewidth, min_val, max_val, py_type, str name, packer_type, bint should_check):
        self.bytewidth = bytewidth
        self.min_val = min_val
        self.max_val = max_val
        self.py_type = py_type
        self.name = name
        self.packer_type = packer_type
        self.should_check = should_check


def float32_to_uint32(n):
    packed = struct.pack("<1f", n)
    (converted,) = struct.unpack("<1L", packed)
    return converted


def uint32_to_float32(n):
    packed = struct.pack("<1L", n)
    (unpacked,) = struct.unpack("<1f", packed)
    return unpacked


def float64_to_uint64(n):
    packed = struct.pack("<1d", n)
    (converted,) = struct.unpack("<1Q", packed)
    return converted


def uint64_to_float64(n):
    packed = struct.pack("<1Q", n)
    (unpacked,) = struct.unpack("<1d", packed)
    return unpacked


def to_numpy_type(number_type):
    if np is not None:
        return np.dtype(number_type.name).newbyteorder('<')
    else:
        raise NumpyRequiredForThisFeature('Numpy was not found.')



#cython: language_level=3,
# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from libc.string cimport memcmp, memcpy, memset
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t, int8_t, int16_t, int32_t, int64_t
from cpython.mem cimport PyMem_Malloc, PyMem_Free

import struct

from .number_types cimport Flags, enforce_number

from flatbuffers import encode

cdef struct FlatbufferType:
    uint8_t bytewidth


cdef struct StoredVtable:
    uint32_t num_fields
    unsigned char* vtable
    int64_t offset        # Offset from the end of the file.
    StoredVtable* next


cdef FlatbufferType Fb_bool_t = FlatbufferType(1)
cdef FlatbufferType Fb_uint8_t = FlatbufferType(1)
cdef FlatbufferType Fb_uint16_t = FlatbufferType(2)
cdef FlatbufferType Fb_uint32_t = FlatbufferType(4)
cdef FlatbufferType Fb_uint64_t = FlatbufferType(8)
cdef FlatbufferType Fb_int8_t = FlatbufferType(1)
cdef FlatbufferType Fb_int16_t = FlatbufferType(2)
cdef FlatbufferType Fb_int32_t = FlatbufferType(4)
cdef FlatbufferType Fb_int64_t = FlatbufferType(8)
cdef FlatbufferType Fb_float32_t = FlatbufferType(4)
cdef FlatbufferType Fb_float64_t = FlatbufferType(8)


cdef inline void Write(packer_type, buf, Py_ssize_t head, Py_ssize_t n):
    """ Write encodes `n` at buf[head] using `packer_type`. """
    packer_type.pack_into(buf, head, n)


cpdef inline void writeBool(bint n, unsigned char* buffer, Py_ssize_t head):
    buffer[head] = 1 if n else 0


cpdef inline void writeUint8(uint8_t n, unsigned char* buffer, Py_ssize_t head):
    buffer[head] = n & 0xFF


cpdef inline void writeUint16(uint16_t n, unsigned char* buffer, Py_ssize_t head):
    buffer[head+1] = (n >> 8) & 0xFF
    buffer[head] = n & 0xFF


cpdef inline void writeUint32(uint32_t n, unsigned char* buffer, Py_ssize_t head):
    buffer[head+3] = (n >> 24) & 0xFF
    buffer[head+2] = (n >> 16) & 0xFF
    buffer[head+1] = (n >> 8) & 0xFF
    buffer[head] = n & 0xFF


cpdef inline void writeUint64(uint64_t n, unsigned char* buffer, Py_ssize_t head):
    buffer[head+7] = (n >> 56) & 0xFF
    buffer[head+6] = (n >> 48) & 0xFF
    buffer[head+5] = (n >> 40) & 0xFF
    buffer[head+4] = (n >> 32) & 0xFF
    buffer[head+3] = (n >> 24) & 0xFF
    buffer[head+2] = (n >> 16) & 0xFF
    buffer[head+1] = (n >> 8) & 0xFF
    buffer[head] = n & 0xFF


cpdef inline void writeInt8(int8_t n, unsigned char* buffer, Py_ssize_t head):
    buffer[head] = n & 0xFF


cpdef inline void writeInt16(int16_t n, unsigned char* buffer, Py_ssize_t head):
    buffer[head+1] = (n >> 8) & 0xFF
    buffer[head] = n & 0xFF


cpdef inline void writeInt32(int32_t n, unsigned char* buffer, Py_ssize_t head):
    buffer[head+3] = (n >> 24) & 0xFF
    buffer[head+2] = (n >> 16) & 0xFF
    buffer[head+1] = (n >> 8) & 0xFF
    buffer[head] = n & 0xFF


cpdef inline void writeInt64(int64_t n, unsigned char* buffer, Py_ssize_t head):
    buffer[head+7] = (n >> 56) & 0xFF
    buffer[head+6] = (n >> 48) & 0xFF
    buffer[head+5] = (n >> 40) & 0xFF
    buffer[head+4] = (n >> 32) & 0xFF
    buffer[head+3] = (n >> 24) & 0xFF
    buffer[head+2] = (n >> 16) & 0xFF
    buffer[head+1] = (n >> 8) & 0xFF
    buffer[head] = n & 0xFF


cdef union double2bytes:
    double d
    unsigned char b[8]


cpdef inline void writeFloat64(double n, unsigned char* buffer, Py_ssize_t head):
    cdef double2bytes t
    t.d = n
    buffer[head+7] = t.b[7]
    buffer[head+6] = t.b[6]
    buffer[head+5] = t.b[5]
    buffer[head+4] = t.b[4]
    buffer[head+3] = t.b[3]
    buffer[head+2] = t.b[2]
    buffer[head+1] = t.b[1]
    buffer[head] = t.b[0]


cdef union float2bytes:
    float f
    unsigned char b[8]


cpdef inline void writeFloat32(float n, unsigned char* buffer, Py_ssize_t head):
    cdef float2bytes t
    t.f = n
    buffer[head+3] = t.b[3]
    buffer[head+2] = t.b[2]
    buffer[head+1] = t.b[1]
    buffer[head] = t.b[0]


class OffsetArithmeticError(RuntimeError):
    """
    Error caused by an Offset arithmetic error. Probably caused by bad
    writing of fields. This is considered an unreachable situation in
    normal circumstances.
    """
    pass


class IsNotNestedError(RuntimeError):
    """
    Error caused by using a Builder to write Object data when not inside
    an Object.
    """
    pass


class IsNestedError(RuntimeError):
    """
    Error caused by using a Builder to begin an Object when an Object is
    already being built.
    """
    pass


class StructIsNotInlineError(RuntimeError):
    """
    Error caused by using a Builder to write a Struct at a location that
    is not the current Offset.
    """
    pass


class BuilderSizeError(RuntimeError):
    """
    Error caused by causing a Builder to exceed the hardcoded limit of 2
    gigabytes.
    """
    pass

class BuilderNotFinishedError(RuntimeError):
    """
    Error caused by not calling `Finish` before calling `Output`.
    """
    pass


# VtableMetadataFields is the count of metadata fields in each vtable.
cdef unsigned int VtableMetadataFields = 2
cdef unsigned int MAX_BUFFER_SIZE = 2**31

cdef class Builder(object):
    """ A Builder is used to construct one or more FlatBuffers.

    Typically, Builder objects will be used from code generated by the `flatc`
    compiler.

    A Builder constructs byte buffers in a last-first manner for simplicity and
    performance during reading.

    Internally, a Builder is a state machine for creating FlatBuffer objects.

    It holds the following internal state:
        - Bytes: an array of bytes.
        - current_vtable: a list of integers.
        - vtables: a list of vtable entries (i.e. a list of list of integers).

    Attributes:
      Bytes: The internal `bytearray` for the Builder.
      finished: A boolean determining if the Builder has been finalized.
    """
    cdef unsigned char* buffer
    cdef Py_ssize_t buffer_length
    cdef Py_ssize_t* current_vtable
    cdef Py_ssize_t current_vtable_length
    cdef Py_ssize_t head
    cdef Py_ssize_t minalign
    cdef int objectEnd
    cdef StoredVtable* vtables
    cdef bint nested
    cdef bint finished
    cdef Py_ssize_t vectorNumElems

    def __init__(self, Py_ssize_t initialSize):
        """Initializes a Builder of size `initial_size`.

        The internal buffer is grown as needed.
        """

        if not (0 <= initialSize <= MAX_BUFFER_SIZE):
            msg = "flatbuffers: Cannot create Builder larger than 2 gigabytes."
            raise BuilderSizeError(msg)
        self.buffer = <unsigned char*> PyMem_Malloc(sizeof(unsigned char) * initialSize)
        if self.buffer is NULL:
            raise MemoryError()
        memset(self.buffer, 0, sizeof(unsigned char) * initialSize)
        self.buffer_length = initialSize
        self.current_vtable = NULL
        self.current_vtable_length = 0
        self.head = initialSize
        self.minalign = 1
        self.objectEnd = -1
        self.nested = False
        self.finished = False
        self.vtables = NULL
        self.vectorNumElems = 0

    def __dealloc__(self):
        PyMem_Free(self.buffer)
        if self.current_vtable != NULL:
            PyMem_Free(self.current_vtable)
            self.current_vtable = NULL
            self.current_vtable_length = 0
        while self.vtables is not NULL:
            next = self.vtables.next
            PyMem_Free(self.vtables.vtable)
            PyMem_Free(self.vtables)
            self.vtables = next

    def Output(self):
        """Return the portion of the buffer that has been used for writing data.

        This is the typical way to access the FlatBuffer data inside the
        builder. If you try to access `Builder.Bytes` directly, you would need
        to manually index it with `Head()`, since the buffer is constructed
        backwards.

        It raises BuilderNotFinishedError if the buffer has not been finished
        with `Finish`.
        """

        if not self.finished:
            raise BuilderNotFinishedError()

        cdef bytes res = self.buffer[self.head:self.buffer_length]

        return res

    @property
    def Bytes(self):
        return self.buffer[0:self.buffer_length]

    def StartObject(self, int numfields):
        """StartObject initializes bookkeeping for writing a new object."""

        self.assertNotNested()

        if numfields > 0:
            self.current_vtable = <Py_ssize_t*>PyMem_Malloc(sizeof(Py_ssize_t) * numfields)
            if self.current_vtable is NULL:
                raise MemoryError()
            memset(self.current_vtable, 0, sizeof(Py_ssize_t) * numfields)
        self.current_vtable_length = numfields
        self.objectEnd = self.Offset()
        self.nested = True

    cdef Py_ssize_t WriteVtable(self):
        """
        WriteVtable serializes the vtable for the current object, if needed.

        Before writing out the vtable, this checks pre-existing vtables for
        equality to this one. If an equal vtable is found, point the object to
        the existing vtable and return.

        Because vtable values are sensitive to alignment of object data, not
        all logically-equal vtables will be deduplicated.

        A vtable has the following format:
          <VOffsetT: size of the vtable in bytes, including this value>
          <VOffsetT: size of the object in bytes, including the vtable offset>
          <VOffsetT: offset for a field> * N, where N is the number of fields
                     in the schema for this type. Includes deprecated fields.
        Thus, a vtable is made of 2 + N elements, each VOffsetT bytes wide.

        An object has the following format:
          <SOffsetT: offset to this object's vtable (may be negative)>
          <byte: data>+
        """
        cdef Py_ssize_t objectOffset, i, off, objectSize, objectStart, vBytes
        cdef Py_ssize_t effective_vtable_length, vtable_offset
        cdef StoredVtable* svt

        # Prepend a zero scalar to the object. Later in this function we'll
        # write an offset here that points to the object's vtable:
        self.PrependSOffsetTRelative(0)  # 32-bit value.

        objectOffset = self.Offset()

        # Trim trailing 0 offsets.
        effective_vtable_length = 0
        for i in range(self.current_vtable_length, 0, -1):
            if self.current_vtable[i-1] != 0:
                effective_vtable_length = i
                break

        # Write out the current vtable in reverse, because
        # serialization occurs in last-first order:
        i = effective_vtable_length - 1
        cdef Py_ssize_t norm_vtable_length = sizeof(uint16_t) * effective_vtable_length
        norm_vtable = <unsigned char*>PyMem_Malloc(norm_vtable_length)
        while i >= 0:
            off = 0
            if self.current_vtable[i] != 0:
                # Forward reference to field;
                # use 32bit number to ensure no overflow:
                off = objectOffset - self.current_vtable[i]

            writeUint16(off, norm_vtable, i*2)
            i -= 1

        # Check if this vtable has been seen already.
        svt = self.vtables
        while svt is not NULL:
            if (svt.num_fields == effective_vtable_length and
                memcmp(norm_vtable, svt.vtable, norm_vtable_length) == 0):
                break
            svt = svt.next

        if svt is not NULL:
            PyMem_Free(self.current_vtable)
            self.current_vtable = NULL
            self.current_vtable_length = 0
            PyMem_Free(norm_vtable)
            objectStart = self.buffer_length - objectOffset
            self.head = objectStart

            # Write the offset to the found vtable in the
            # already-allocated SOffsetT at the beginning of this object.
            # The cached vtable contains its offset from the end of the buffer.
            vtable_offset = svt.offset
            writeInt32(vtable_offset - objectOffset, self.buffer, objectStart)
            return self.Offset()

        i = effective_vtable_length - 1
        while i >= 0:
            off = 0
            if self.current_vtable[i] != 0:
                # Forward reference to field;
                # use 32bit number to ensure no overflow:
                off = objectOffset - self.current_vtable[i]

            self.PrependVOffsetT(off)
            i -= 1
        # The two metadata fields are written last.

        # First, store the object bytesize:
        objectSize = objectOffset - self.objectEnd
        self.PrependVOffsetT(objectSize)

        # Second, store the vtable bytesize:
        vBytes = effective_vtable_length + VtableMetadataFields
        vBytes *= Fb_uint16_t.bytewidth
        self.PrependVOffsetT(vBytes)

        # Next, write the offset to the new vtable in the
        # already-allocated SOffsetT at the beginning of this object:
        objectStart = self.buffer_length - objectOffset
        writeUint32(self.Offset() - objectOffset, self.buffer, objectStart)

        # Finally, store this vtable in memory for future
        # deduplication:
        svt = <StoredVtable*>PyMem_Malloc(sizeof(StoredVtable))
        svt.num_fields = effective_vtable_length
        svt.vtable = norm_vtable
        svt.next = self.vtables
        svt.offset = self.Offset()
        self.vtables = svt

        PyMem_Free(self.current_vtable)
        self.current_vtable = NULL
        self.current_vtable_length = 0
        return objectOffset

    def EndObject(self):
        """EndObject writes data necessary to finish object construction."""
        self.assertNested()
        self.nested = False
        return self.WriteVtable()

    cdef void growByteBuffer(self):
        """Doubles the size of the byteslice, and copies the old data towards
           the end of the new buffer (since we build the buffer backwards)."""
        cdef Py_ssize_t newSize
        if self.buffer_length == MAX_BUFFER_SIZE:
            msg = "flatbuffers: cannot grow buffer beyond 2 gigabytes"
            raise BuilderSizeError(msg)

        newSize = min(self.buffer_length * 2, MAX_BUFFER_SIZE)
        if newSize == 0:
            newSize = 1
        buffer_2 = <unsigned char *>PyMem_Malloc(sizeof(unsigned char) * newSize)
        memset(buffer_2, 0, sizeof(unsigned char) * newSize)
        memcpy(&buffer_2[newSize-self.buffer_length], self.buffer, self.buffer_length)
        self.buffer_length = newSize
        PyMem_Free(self.buffer)
        self.buffer = buffer_2

    cdef Py_ssize_t Offset(self):
        """Offset relative to the end of the buffer."""
        return self.buffer_length - self.head

    cdef Pad(self, Py_ssize_t n):
        """Pad places zeros at the current offset."""
        cdef int i
        for i in range(n):
            self.Place(0, &Fb_uint8_t)

    cdef void Prep(self, Py_ssize_t size, Py_ssize_t additionalBytes):
        """
        Prep prepares to write an element of `size` after `additional_bytes`
        have been written, e.g. if you write a string, you need to align
        such the int length field is aligned to SizeInt32, and the string
        data follows it directly.
        If all you need to do is align, `additionalBytes` will be 0.
        """
        cdef long alignSize, updated_head, oldBufSize
        # Track the biggest thing we've ever aligned to.
        if size > self.minalign:
            self.minalign = size

        # Find the amount of alignment needed such that `size` is properly
        # aligned after `additionalBytes`:
        alignSize = (~(self.buffer_length - self.head + additionalBytes)) + 1
        alignSize &= (size - 1)

        # Reallocate the buffer if needed:
        while self.head < alignSize+size+additionalBytes:
            oldBufSize = self.buffer_length
            self.growByteBuffer()
            updated_head = self.head + self.buffer_length - oldBufSize
            self.head = updated_head
        self.Pad(alignSize)

    cdef void PrependSOffsetTRelative(self, int off):
        """
        PrependSOffsetTRelative prepends an SOffsetT, relative to where it
        will be written.
        """

        # Ensure alignment is already done:
        self.Prep(Fb_int32_t.bytewidth, 0)
        if not (off <= self.Offset()):
            msg = "flatbuffers: Offset arithmetic error."
            raise OffsetArithmeticError(msg)
        off2 = self.Offset() - off + Fb_int32_t.bytewidth
        self.PlaceSOffsetT(off2)

    cpdef void PrependUOffsetTRelative(self, Py_ssize_t off):
        """Prepends an unsigned offset into vector data, relative to where it
        will be written.
        """

        # Ensure alignment is already done:
        self.Prep(Fb_uint32_t.bytewidth, 0)
        if not (off <= self.Offset()):
            msg = "flatbuffers: Offset arithmetic error."
            raise OffsetArithmeticError(msg)
        off2 = self.Offset() - off + Fb_uint32_t.bytewidth
        self.PlaceUOffsetT(off2)

    def StartVector(self, int elemSize, Py_ssize_t numElems, Py_ssize_t alignment):
        """
        StartVector initializes bookkeeping for writing a new vector.

        A vector has the following format:
          - <UOffsetT: number of elements in this vector>
          - <T: data>+, where T is the type of elements of this vector.
        """

        self.assertNotNested()
        self.nested = True
        self.vectorNumElems = numElems
        self.Prep(Fb_uint32_t.bytewidth, elemSize*numElems)
        self.Prep(alignment, elemSize*numElems)  # In case alignment > int.
        return self.Offset()

    cpdef Py_ssize_t EndVector(self):
        """EndVector writes data necessary to finish vector construction."""

        self.assertNested()
        ## @cond FLATBUFFERS_INTERNAL
        self.nested = False
        ## @endcond
        # we already made space for this, so write without PrependUint32
        self.PlaceUOffsetT(self.vectorNumElems)
        return self.Offset()

    cpdef Py_ssize_t CreateString(self, str s, encoding='utf-8', errors='strict'):
        """CreateString writes a null-terminated byte string as a vector."""
        cdef bytes x
        cdef Py_ssize_t l

        self.assertNotNested()
        ## @cond FLATBUFFERS_INTERNAL
        self.nested = True
        ## @endcond

        x = s.encode(encoding)
        l = len(x)

        self.Prep(Fb_uint32_t.bytewidth, (l+1)*Fb_uint8_t.bytewidth)
        self.Place(0, &Fb_uint8_t)

        self.head = self.head - l

        cdef char* c_string = x

        memcpy(<void*>&self.buffer[self.head], c_string, l)
        self.vectorNumElems = l

        return self.EndVector()

    cpdef Py_ssize_t CreateByteVector(self, bytes x):
        """CreateString writes a byte vector."""
        self.assertNotNested()
        self.nested = True

        cdef l = len(x)

        self.Prep(Fb_uint32_t.bytewidth, l*Fb_uint8_t.bytewidth)

        self.head = self.head - l

        cdef char* c_string = x

        memcpy(<void*>&self.buffer[self.head], c_string, l)
        self.vectorNumElems = l

        return self.EndVector()

    cdef void assertNested(self) except *:
        """
        Check that we are in the process of building an object.
        """

        if not self.nested:
            raise IsNotNestedError()

    cdef void assertNotNested(self) except *:
        """
        Check that no other objects are being built while making this
        object. If not, raise an exception.
        """

        if self.nested:
            raise IsNestedError()

    cdef void assertStructIsInline(self, Py_ssize_t obj) except *:
        """
        Structs are always stored inline, so need to be created right
        where they are used. You'll get this error if you created it
        elsewhere.
        """

        if obj != self.Offset():
            msg = ("flatbuffers: Tried to write a Struct at an Offset that "
                   "is different from the current Offset of the Builder.")
            raise StructIsNotInlineError(msg)

    cdef void Slot(self, Py_ssize_t slotnum) except *:
        """
        Slot sets the vtable key `voffset` to the current location in the
        buffer.

        """
        self.assertNested()
        if slotnum >= self.current_vtable_length:
            raise IndexError()
        self.current_vtable[slotnum] = self.Offset()

    cdef Py_ssize_t __Finish(self, Py_ssize_t rootTable, bint sizePrefix):
        """Finish finalizes a buffer, pointing to the given `rootTable`."""
        cdef Py_ssize_t prepSize, size

        prepSize = Fb_uint32_t.bytewidth
        if sizePrefix:
            prepSize += Fb_int32_t.bytewidth

        self.Prep(self.minalign, prepSize)
        self.PrependUOffsetTRelative(rootTable)
        if sizePrefix:
            size = self.buffer_length - self.head
            self.PrependInt32(size)
        self.finished = True
        return self.head

    def Finish(self, int rootTable):
        """Finish finalizes a buffer, pointing to the given `rootTable`."""
        return self.__Finish(rootTable, False)

    def FinishSizePrefixed(self, Py_ssize_t rootTable):
        """
        Finish finalizes a buffer, pointing to the given `rootTable`,
        with the size prefixed.
        """
        return self.__Finish(rootTable, True)

    cdef void Prepend(self, FlatbufferType* type, int64_t off):
        self.Prep(type.bytewidth, 0)
        self.Place(off, type)

    cdef void PrependSlot(self, FlatbufferType* type, Py_ssize_t o, int64_t x, int64_t d):
        if x != d:
            self.Prepend(type, x)
            self.Slot(o)

    def PrependBoolSlot(self, Py_ssize_t o, bint x, bint d):
        self.PrependSlot(&Fb_bool_t, o, x, d)

    def PrependByteSlot(self, Py_ssize_t o, int x, int d):
        self.PrependSlot(&Fb_uint8_t, o, x, d)

    def PrependUint8Slot(self, Py_ssize_t o, uint8_t x, uint8_t d):
        self.PrependSlot(&Fb_uint8_t, o, x, d)

    def PrependUint16Slot(self, Py_ssize_t o, uint16_t x, uint16_t d):
        self.PrependSlot(&Fb_uint16_t, o, x, d)

    def PrependUint32Slot(self, Py_ssize_t o, uint32_t x, uint32_t d):
        self.PrependSlot(&Fb_uint32_t, o, x, d)

    def PrependUint64Slot(self, Py_ssize_t o, uint64_t x, uint64_t d):
        self.PrependSlot(&Fb_uint64_t, o, x, d)

    def PrependInt8Slot(self, Py_ssize_t o, int8_t x, int8_t d):
        self.PrependSlot(&Fb_int8_t, o, x, d)

    def PrependInt16Slot(self, Py_ssize_t o, int16_t x, int16_t d):
        self.PrependSlot(&Fb_int16_t, o, x, d)

    def PrependInt32Slot(self, Py_ssize_t o, int32_t x, int32_t d):
        self.PrependSlot(&Fb_int32_t, o, x, d)

    def PrependInt64Slot(self, Py_ssize_t o, int64_t x, int64_t d):
        self.PrependSlot(&Fb_int64_t, o, x, d)

    def PrependFloat32Slot(self, Py_ssize_t o, float x, float d):
        if x != d:
            self.Prep(Fb_float32_t.bytewidth, 0)
            self.head = self.head - Fb_float32_t.bytewidth
            writeFloat32(x, self.buffer, self.head)
            self.Slot(o)

    def PrependFloat64Slot(self, Py_ssize_t o, double x, double d):
        if x != d:
            self.Prep(Fb_float64_t.bytewidth, 0)
            self.head = self.head - Fb_float64_t.bytewidth
            writeFloat64(x, self.buffer, self.head)
            self.Slot(o)

    def PrependUOffsetTRelativeSlot(self, Py_ssize_t o, Py_ssize_t x, Py_ssize_t d):
        """
        PrependUOffsetTRelativeSlot prepends an UOffsetT onto the object at
        vtable slot `o`. If value `x` equals default `d`, then the slot will
        be set to zero and no other data will be written.
        """
        if x != d:
            self.PrependUOffsetTRelative(x)
            self.Slot(o)

    cdef PrependStructSlot(self, v, x, d):
        """
        PrependStructSlot prepends a struct onto the object at vtable slot `o`.
        Structs are stored inline, so nothing additional is being added.
        In generated code, `d` is always 0.
        """
        if x != d:
            self.assertStructIsInline(x)
            self.Slot(v)

    def PrependBool(self, x):
        """Prepend a `bool` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prepend(&Fb_bool_t, x)

    def PrependByte(self, x):
        """Prepend a `byte` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prepend(&Fb_uint8_t, x)

    def PrependUint8(self, x):
        """Prepend an `uint8` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prepend(&Fb_uint8_t, x)

    def PrependUint16(self, x):
        """Prepend an `uint16` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prepend(&Fb_uint16_t, x)

    def PrependUint32(self, uint32_t x):
        """Prepend an `uint32` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prepend(&Fb_uint32_t, x)

    def PrependUint64(self, x):
        """Prepend an `uint64` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prep(Fb_uint64_t.bytewidth, 0)
        self.Place(x, &Fb_uint64_t)

    def PrependInt8(self, x):
        """Prepend an `int8` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prepend(&Fb_int8_t, x)

    def PrependInt16(self, x):
        """Prepend an `int16` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prepend(&Fb_int16_t, x)

    cpdef PrependInt32(self, int32_t x):
        """Prepend an `int32` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prepend(&Fb_int32_t, x)

    def PrependInt64(self, x):
        """Prepend an `int64` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prepend(&Fb_int64_t, x)

    def PrependFloat32(self, x):
        """Prepend a `float32` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prep(Fb_float32_t.bytewidth, 0)
        self.head = self.head - Fb_float32_t.bytewidth
        writeFloat32(x, self.buffer, self.head)

    def PrependFloat64(self, x):
        """Prepend a `float64` to the Builder buffer.

        Note: aligns and checks for space.
        """
        self.Prep(Fb_float64_t.bytewidth, 0)
        self.head = self.head - Fb_float64_t.bytewidth
        writeFloat64(x, self.buffer, self.head)

    cdef void PrependVOffsetT(self, uint16_t x):
        self.Prepend(&Fb_uint16_t, x)

    cdef void Place(self, uint64_t x, FlatbufferType *fb_type):
        """
        Place prepends a value specified by `flags` to the Builder,
        without checking for available space.
        """
        self.head = self.head - fb_type.bytewidth
        #Write(flags.packer_type, self.Bytes, self.head, x)

        if fb_type == &Fb_bool_t:
            writeBool(x, self.buffer, self.head)
        elif fb_type == &Fb_uint8_t:
            writeUint8(x, self.buffer, self.head)
        elif fb_type == &Fb_uint16_t:
            writeUint16(x, self.buffer, self.head)
        elif fb_type == &Fb_uint32_t:
            writeUint32(x, self.buffer, self.head)
        elif fb_type == &Fb_uint64_t:
            writeUint64(x, self.buffer, self.head)
        elif fb_type == &Fb_int8_t:
            writeInt8(x, self.buffer, self.head)
        elif fb_type == &Fb_int16_t:
            writeInt16(x, self.buffer, self.head)
        elif fb_type == &Fb_int32_t:
            writeInt32(x, self.buffer, self.head)
        elif fb_type == &Fb_int64_t:
            writeInt64(x, self.buffer, self.head)

    cdef void PlaceVOffsetT(self, uint16_t x):
        """PlaceVOffsetT prepends a VOffsetT to the Builder, without checking
        for space.
        """
        self.head = self.head - Fb_uint16_t.bytewidth
        writeUint16(x, self.buffer, self.head)

    cdef void PlaceSOffsetT(self, int32_t x):
        """PlaceSOffsetT prepends a SOffsetT to the Builder, without checking
        for space.
        """
        self.head = self.head - Fb_int32_t.bytewidth
        #Write(soffset, self.Bytes, self.head, x)
        writeInt32(x, self.buffer, self.head)

    cdef void PlaceUOffsetT(self, uint32_t x):
        """PlaceUOffsetT prepends a UOffsetT to the Builder, without checking
        for space.
        """
        self.head = self.head - Fb_uint32_t.bytewidth
        #Write(uoffset, self.Bytes, self.head, x)
        writeUint32(x, self.buffer, self.head)



from mypy.plugin import Plugin
from mypy.plugins.attrs import attr_class_makers

# This work just like attr.s i.e. it has the same defaults as attr.s
attr_class_makers.add("flattr.Flatbuffer")
attr_class_makers.add("flattr.from_package")


class MyPlugin(Plugin):
    # Our plugin does nothing but it has to exist so this file gets loaded.
    pass


def plugin(version):
    return MyPlugin



import hashlib
import linecache
from collections.abc import Sequence
from enum import Enum, IntEnum, unique
from importlib import import_module
from re import sub
from sys import modules
from types import GenericAlias
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    _GenericAlias,
)

import attr
from attr import fields, has
from flatbuffers.number_types import (
    BoolFlags,
    Float32Flags,
    Float64Flags,
    Int8Flags,
    Int16Flags,
    Int32Flags,
    Int64Flags,
    Uint8Flags,
    Uint16Flags,
    Uint32Flags,
    Uint64Flags,
)

try:
    from .cflattr.builder import Builder
except ImportError:
    from flatbuffers.builder import Builder


_T = TypeVar("_T")


def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_descriptors: Tuple[Union[type, Callable[..., Any]], ...] = (
        (attr.attrib, attr.field)
    ),
) -> Callable[[_T], _T]:
    return lambda c: c


UNION_CL = "__fb_union_cl"


@__dataclass_transform__()
def Flatbuffer(fb_cl, frozen: bool = False, repr: bool = True):
    def wrapper(cl):
        res = attr.s(slots=True, frozen=frozen, repr=repr)(cl)
        res.__fb_module__ = modules[fb_cl.__module__]
        res.__fb_class__ = fb_cl
        _make_fb_functions(res)

        return res

    return wrapper


def FlatbufferEnum(fb_cl):
    def wrapper(cl):
        res = unique(cl)
        res.__fb_module__ = modules[fb_cl.__module__]
        res.__fb_class__ = fb_cl

        if cl.__name__ != fb_cl.__name__:
            raise ValueError(f"Names don't match: {cl.__name__}/{fb_cl.__name__}.")

        for member in cl:
            if not hasattr(fb_cl, member.name):
                raise ValueError(f"{cl}/{member.name} doesn't match {fb_cl}.")

        return res

    return wrapper


@__dataclass_transform__()
def from_package(pkg, frozen: bool = False, repr: bool = True):
    def wrap_cls(cl):
        cl_name = cl.__name__
        module = import_module(f"{pkg.__package__}.{cl_name}")
        fb_cl = getattr(module, cl_name)
        return Flatbuffer(fb_cl, frozen=frozen, repr=repr)(cl)

    return wrap_cls


def from_package_enum(pkg):
    def wrap_cls(cl):
        cl_name = cl.__name__
        module = import_module(f"{pkg.__package__}.{cl_name}")
        fb_cl = getattr(module, cl_name)
        return FlatbufferEnum(fb_cl)(cl)

    return wrap_cls


Flatbuffer.from_package = from_package
FlatbufferEnum.from_package = from_package_enum


none_type = type(None)


def is_generic_subclass(t, s):
    return (isinstance(t, _GenericAlias) and t.__origin__ is s) or (
        isinstance(t, GenericAlias) and t.__origin__ is s
    )


def _make_fb_functions(cl):
    """Inspect the given class for any non-nestable buffers.

    Non-nestables are other tables, strings and lists of non-nestables.
    """
    fn_name = "__fb_nonnestables__"
    strings = []
    optional_strings = []
    byte_fields = []
    optional_bytes = []
    tables = []
    optional_tables = []
    lists_of_tables: List[Tuple[str, Type, bool]] = []
    seqs_of_tables: List[Tuple[str, Type, bool]] = []
    lists_of_strings: List[Tuple[str, bool]] = []
    seqs_of_strings: List[Tuple[str, bool]] = []
    lists_of_scalars: List[Tuple[str, Type, Any, bool]] = []
    seqs_of_scalars: List[Tuple[str, Type, Any, bool]] = []
    lists_of_enums: List[Tuple[str, Type, Any, bool]] = []
    seqs_of_enums: List[Tuple[str, Type, Any, bool]] = []
    enums = []
    inlines = []
    unions: Tuple[str, List[Type], Any] = []
    for field in fields(cl):
        type = field.type
        if type is str:
            strings.append(field.name)
        elif type in (bytes, bytearray):
            byte_fields.append(field.name)
        elif has(type):
            tables.append(field.name)
        elif getattr(type, "__origin__", None) is Union:
            union_args = type.__args__
            if len(union_args) == 2 and none_type in union_args:
                # This is an optional field.
                if union_args[0] is str:
                    optional_strings.append(field.name)
                elif union_args[0] is bytes:
                    optional_bytes.append(field.name)
                elif has(union_args[0]):
                    optional_tables.append(field.name)
                elif is_generic_subclass(union_args[0], list):
                    arg = union_args[0].__args__[0]
                    if arg is str:
                        lists_of_strings.append((field.name, True))
                    elif has(arg):
                        lists_of_tables.append((field.name, arg, True))
                    elif arg in (int, float, bool):
                        lists_of_scalars.append(
                            (
                                field.name,
                                arg,
                                _get_scalar_list_type(cl, field.name)[0],
                                True,
                            )
                        )
                    elif issubclass(arg, IntEnum):
                        lists_of_enums.append(
                            (
                                field.name,
                                arg,
                                _get_scalar_list_type(cl, field.name)[0],
                                True,
                            )
                        )
                elif is_generic_subclass(union_args[0], Sequence):
                    arg = union_args[0].__args__[0]
                    if arg is str:
                        seqs_of_strings.append((field.name, True))
                    elif has(arg):
                        seqs_of_tables.append((field.name, arg, True))
                    elif arg in (int, float, bool):
                        seqs_of_scalars.append(
                            (
                                field.name,
                                arg,
                                _get_scalar_list_type(cl, field.name)[0],
                                True,
                            )
                        )
                    elif issubclass(arg, IntEnum):
                        seqs_of_enums.append(
                            (
                                field.name,
                                arg,
                                _get_scalar_list_type(cl, field.name)[0],
                                True,
                            )
                        )
                elif issubclass(union_args[0], IntEnum):
                    raise ValueError("Flatbuffers don't support optional enums.")
            else:
                unions.append((field.name, type.__args__, field.metadata[UNION_CL]))
        elif is_generic_subclass(type, list):
            arg = type.__args__[0]
            if arg is str:
                lists_of_strings.append((field.name, False))
            elif has(arg):
                lists_of_tables.append((field.name, arg, False))
            elif arg in (int, float, bool):
                lists_of_scalars.append(
                    (field.name, arg, _get_scalar_list_type(cl, field.name)[0], False)
                )
            elif issubclass(arg, IntEnum):
                lists_of_enums.append(
                    (field.name, arg, _get_scalar_list_type(cl, field.name)[0], False)
                )
        elif is_generic_subclass(type, Sequence):
            arg = type.__args__[0]
            if arg is str:
                seqs_of_strings.append((field.name, False))
            elif has(arg):
                seqs_of_tables.append((field.name, arg, False))
            elif arg in (int, float, bool):
                seqs_of_scalars.append(
                    (field.name, arg, _get_scalar_list_type(cl, field.name)[0], False)
                )
            elif issubclass(arg, IntEnum):
                seqs_of_enums.append(
                    (field.name, arg, _get_scalar_list_type(cl, field.name)[0], False)
                )
        elif issubclass(type, IntEnum):
            enums.append(field.name)
        else:
            inlines.append(field.name)

    setattr(
        cl,
        fn_name,
        _make_nonnestables_fn(
            cl,
            strings,
            optional_strings,
            byte_fields,
            tables,
            optional_tables,
            lists_of_tables + seqs_of_tables,
            lists_of_strings + seqs_of_strings,
            unions,
        ),
    )

    setattr(
        cl,
        "__fb_add_to_builder__",
        _make_add_to_builder_fn(
            cl,
            strings,
            optional_strings,
            byte_fields,
            optional_bytes,
            tables,
            optional_tables,
            lists_of_tables + seqs_of_tables,
            lists_of_strings + seqs_of_strings,
            lists_of_scalars + lists_of_enums + seqs_of_scalars + seqs_of_enums,
            inlines + enums,
            unions,
        ),
    )

    setattr(cl, "__fb_from_bytes__", _make_from_bytes_fn(cl))
    setattr(
        cl,
        "__fb_from_fb__",
        _make_from_fb_fn(
            cl,
            strings,
            optional_strings,
            byte_fields,
            optional_bytes,
            enums,
            tables,
            optional_tables,
            lists_of_tables,
            lists_of_strings,
            unions,
            inlines,
            lists_of_scalars,
            lists_of_enums,
            seqs_of_tables,
            seqs_of_scalars,
            seqs_of_enums,
            seqs_of_strings,
        ),
    )


def model_to_bytes(inst, builder: Optional[Builder] = None) -> bytes:
    builder = Builder(10000) if builder is None else builder
    fb_items = inst.__fb_nonnestables__()
    string_offsets = {}
    node_offsets = {}

    for fb_item, fb_type, fb_vec_start in fb_items:
        if fb_type is FBItemType.VECTOR:
            # Make a vector.
            fb_vec_start(builder, len(fb_item))
            for item in reversed(fb_item):
                builder.PrependUOffsetTRelative(node_offsets[id(item)])
            offset = builder.EndVector()
            node_offsets[id(fb_item)] = offset
        else:
            item_id = id(fb_item)
            if item_id not in node_offsets:
                offset = fb_item.__fb_add_to_builder__(
                    builder, string_offsets, node_offsets
                )
                node_offsets[item_id] = offset
    builder.Finish(offset)  # Last offset.
    return bytes(builder.Output())


def model_from_bytes(cl, payload):
    return cl.__fb_from_bytes__(payload)


class FBItemType(str, Enum):
    TABLE = "table"
    VECTOR = "vector"


FBItem = Tuple[Any, FBItemType, Callable]

fb_number_type_to_builder_prepend = {
    BoolFlags: "PrependBool",
    Uint8Flags: "PrependUint8",
    Uint16Flags: "PrependUint16",
    Uint32Flags: "PrependUint32",
    Uint64Flags: "PrependUint64",
    Int8Flags: "PrependInt8",
    Int16Flags: "PrependInt16",
    Int32Flags: "PrependInt32",
    Int64Flags: "PrependInt64",
    Float32Flags: "PrependFloat32",
    Float64Flags: "PrependFloat64",
}


def _make_nonnestables_fn(
    cl,
    string_fields: List[str],
    optional_strings: List[str],
    byte_fields: List[str],
    table_fields: List[str],
    optional_tables: List[str],
    lists_of_tables: List[Tuple[str, Type, bool]],
    lists_of_strings: List[str],
    unions: List[Tuple[str, List[Type], Type]],
) -> Callable[[], Tuple[Set[str], List[bytes], List[FBItem]]]:
    name = cl.__fb_class__.__name__
    mod = cl.__fb_module__
    globs = {"FBTable": FBItemType.TABLE, "FBVector": FBItemType.VECTOR}
    lines = []
    lines.append("def __fb_nonnestables__(self):")

    lines.append("    fb_items = []")

    for table_field in table_fields:
        lines.append(
            f"    {table_field}_items = self.{table_field}.__fb_nonnestables__()"
        )
        lines.append(f"    fb_items.extend({table_field}_items)")

    for table_field, union_classes, _ in unions:
        if none_type not in union_classes:
            lines.append(
                f"    {table_field}_items = self.{table_field}.__fb_nonnestables__()"
            )
            lines.append(f"    fb_items.extend({table_field}_items)")
        else:
            lines.append(f"    if self.{table_field} is not None:")
            lines.append(
                f"        {table_field}_items = self.{table_field}.__fb_nonnestables__()"
            )
            lines.append(f"        fb_items.extend({table_field}_items)")

    for table_field in optional_tables:
        lines.append(f"    if self.{table_field} is not None:")
        lines.append(
            f"        {table_field}_items = self.{table_field}.__fb_nonnestables__()"
        )
        lines.append(f"        fb_items.extend({table_field}_items)")

    for field, _, is_optional in lists_of_tables:
        norm_field_name = _normalize_fn(field)
        globs[f"{field}VectorStart"] = getattr(
            mod, f"{cl.__name__}Start{norm_field_name}Vector"
        )

        lines.append(f"    {field}_items = self.{field}")

        indent = ""
        if is_optional:
            lines.append(f"    if {field}_items is not None:")
            indent = "  "

        lines.append(f"    {indent}for item in {field}_items:")
        lines.append(f"    {indent}    i_is = item.__fb_nonnestables__()")
        lines.append(f"    {indent}    fb_items.extend(i_is)")
        lines.append(f"    {indent}vec_start = {field}VectorStart")
        lines.append(
            f"    {indent}fb_items.append(({field}_items, FBVector, vec_start))"
        )

    lines.append("    fb_items.append((self, FBTable, None))")

    lines.append("    return fb_items")

    sha1 = hashlib.sha1()
    sha1.update(name.encode("utf-8"))
    unique_filename = "<FB nonnestables for %s, %s>" % (name, sha1.hexdigest())
    script = "\n".join(lines)
    eval(compile(script, unique_filename, "exec"), globs)

    linecache.cache[unique_filename] = (
        len(script),
        None,
        script.splitlines(True),
        unique_filename,
    )

    return globs["__fb_nonnestables__"]


def _make_add_to_builder_fn(
    cl,
    string_fields: List[str],
    optional_strings: List[str],
    byte_fields: List[str],
    optional_bytes: List[str],
    table_fields: List[str],
    optional_tables: List[str],
    lists_of_tables: List[Tuple[str, Type, bool]],
    lists_of_strings: List[Tuple[str, bool]],
    lists_of_scalars: List[Tuple[str, Type]],
    inlines: List[str],
    unions: List[Tuple[str, Type, Type]],
):
    name = cl.__fb_class__.__name__
    mod = cl.__fb_module__
    start = getattr(mod, f"{name}Start")
    num_slots = _get_num_slots(start)
    globs = {}
    lines = []
    lines.append("def __fb_add_to_builder__(self, builder, strs, nodes):")

    for field in string_fields:
        lines.append(f"    __fb_self_{field} = self.{field}")
        lines.append(f"    if __fb_self_{field} not in strs:")
        lines.append(
            f"        strs[__fb_self_{field}] = builder.CreateString(__fb_self_{field})"
        )

    for field in optional_strings:
        lines.append(f"    __fb_self_{field} = self.{field}")
        lines.append(
            f"    if __fb_self_{field} is not None and __fb_self_{field} not in strs:"
        )
        lines.append(
            f"        strs[__fb_self_{field}] = builder.CreateString(__fb_self_{field})"
        )

    for field in byte_fields:
        lines.append(f"    __fb_self_{field} = self.{field}")
        lines.append(f"    if id(__fb_self_{field}) not in nodes:")
        lines.append(
            f"        nodes[id(__fb_self_{field})] = builder.CreateByteVector(__fb_self_{field})"
        )

    for field in optional_bytes:
        lines.append(f"    __fb_self_{field} = self.{field}")
        lines.append(
            f"    if __fb_self_{field} is not None and id(__fb_self_{field}) not in nodes:"
        )
        lines.append(
            f"        nodes[id(__fb_self_{field})] = builder.CreateByteVector(__fb_self_{field})"
        )

    for field, is_optional in lists_of_strings:
        norm_field_name = _normalize_fn(field)
        globs[f"{field}StartVector"] = getattr(
            mod, f"{cl.__name__}Start{norm_field_name}Vector"
        )

        lines.append(f"    __fb_self_{field} = self.{field}")
        i = ""
        if is_optional:
            i = "    "
            lines.append(f"    if __fb_self_{field} is not None:")

        lines.append(f"{i}    __fb_self_{field}_offsets = []")
        lines.append(f"{i}    for e in __fb_self_{field}:")
        lines.append(f"{i}        if e in strs:")
        lines.append(f"{i}            __fb_self_{field}_offsets.append(strs[e])")
        lines.append(f"{i}        else:")
        lines.append(f"{i}            offset = builder.CreateString(e)")
        lines.append(f"{i}            strs[e] = offset")
        lines.append(f"{i}            __fb_self_{field}_offsets.append(offset)")
        lines.append(f"{i}    {field}StartVector(builder, len(__fb_self_{field}))")
        lines.append(f"{i}    for o in reversed(__fb_self_{field}_offsets):")
        lines.append(f"{i}        builder.PrependUOffsetTRelative(o)")
        lines.append(f"{i}    __fb_self_{field}_offset = builder.EndVector()")

    for field, _, fb_number_type, is_optional in lists_of_scalars:
        norm_field_name = _normalize_fn(field)
        prepend = fb_number_type_to_builder_prepend[fb_number_type]
        globs[f"{field}StartVector"] = getattr(
            mod, f"{cl.__name__}Start{norm_field_name}Vector"
        )

        lines.append(f"    __fb_self_{field} = self.{field}")
        indent = ""
        if is_optional:
            lines.append(f"    if __fb_self_{field} is not None:")
            indent = "  "
        lines.append(f"    {indent}{field}StartVector(builder, len(__fb_self_{field}))")
        lines.append(f"    {indent}for o in reversed(__fb_self_{field}):")
        lines.append(f"    {indent}    builder.{prepend}(o)")
        lines.append(f"    {indent}__fb_self_{field}_offset = builder.EndVector()")

    lines.append(f"    builder.StartObject({num_slots})")

    for field in string_fields:
        norm_field_name = _normalize_fn(field)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)
        slot_num, default = _get_offsets_for_string(field_start)
        lines.append(
            f"    builder.PrependUOffsetTRelativeSlot({slot_num}, strs[__fb_self_{field}], {default})"
        )

    for field in optional_strings:
        norm_field_name = _normalize_fn(field)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)
        globs[field_starter_name] = field_start
        lines.append(f"    if __fb_self_{field} is not None:")
        lines.append(f"        {field_starter_name}(builder, strs[__fb_self_{field}])")

    for field in byte_fields:
        norm_field_name = _normalize_fn(field)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)
        slot_num, default = _get_offsets_for_string(field_start)
        lines.append(
            f"    builder.PrependUOffsetTRelativeSlot({slot_num}, nodes[id(__fb_self_{field})], {default})"
        )

    for field in optional_bytes:
        norm_field_name = _normalize_fn(field)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)
        slot_num, default = _get_offsets_for_string(field_start)
        lines.append(f"    if __fb_self_{field} is not None:")
        lines.append(
            f"        builder.PrependUOffsetTRelativeSlot({slot_num}, nodes[id(__fb_self_{field})], {default})"
        )

    for field, is_optional in lists_of_strings:
        norm_field_name = _normalize_fn(field)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)
        globs[field_starter_name] = field_start
        prefix = ""
        if is_optional:
            prefix = f"if __fb_self_{field} is not None: "
        lines.append(
            f"    {prefix}{field_starter_name}(builder, __fb_self_{field}_offset)"
        )

    for field in table_fields + byte_fields:
        norm_field_name = _normalize_fn(field)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)
        globs[field_starter_name] = field_start
        lines.append(f"    {field_starter_name}(builder, nodes[id(self.{field})])")

    for field, _, is_optional in lists_of_tables:
        norm_field_name = _normalize_fn(field)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)
        lines.append(f"    __fb_self_{field} = self.{field}")
        globs[field_starter_name] = field_start
        if is_optional:
            prefix = f"if __fb_self_{field} is not None: "
        else:
            prefix = ""
        lines.append(
            f"    {prefix}{field_starter_name}(builder, nodes[id(__fb_self_{field})])"
        )

    for field, _, _, is_optional in lists_of_scalars:
        norm_field_name = _normalize_fn(field)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)
        globs[field_starter_name] = field_start
        if is_optional:
            prefix = f"if __fb_self_{field} is not None: "
        else:
            prefix = ""
        lines.append(
            f"    {prefix}{field_starter_name}(builder, __fb_self_{field}_offset)"
        )

    for field in optional_tables:
        norm_field_name = _normalize_fn(field)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)
        globs[field_starter_name] = field_start
        lines.append(f"    if self.{field} is not None:")
        lines.append(f"        {field_starter_name}(builder, nodes[id(self.{field})])")

    for field in inlines:
        norm_field_name = _normalize_fn(field)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)
        globs[field_starter_name] = field_start
        lines.append(f"    {field_starter_name}(builder, self.{field})")

    for field, union_types, fb_enum in unions:
        norm_field_name = _normalize_fn(field)
        type_adder_name = f"{name}Add{norm_field_name}Type"
        type_adder = getattr(mod, type_adder_name)
        field_starter_name = f"{name}Add{norm_field_name}"
        field_start = getattr(mod, field_starter_name)

        union_dict_name = f"{norm_field_name}_union_types"
        # Unions might be prefixed by a namespace string, depending on how
        # they're defined.
        stripped_union_dict = {k.split("_")[-1]: v for k, v in fb_enum.__dict__.items()}
        union_dict = {
            t: stripped_union_dict[t.__name__]
            for t in union_types
            if t is not none_type
        }
        # Flatbuffer unions always have a special member, NONE, signifying
        # a missing value. If None is part of union_types, we handle it
        # specially.
        if none_type in union_types:
            union_dict[none_type] = stripped_union_dict["NONE"]

        globs[type_adder_name] = type_adder
        globs[field_starter_name] = field_start
        globs[union_dict_name] = union_dict
        lines.append(
            f"    {type_adder_name}(builder, {union_dict_name}[self.{field}.__class__])"
        )
        if none_type in union_types:
            lines.append(f"    if self.{field} is not None:")
            lines.append(
                f"        {field_starter_name}(builder, nodes[id(self.{field})])"
            )
        else:
            lines.append(f"    {field_starter_name}(builder, nodes[id(self.{field})])")

    lines.append("    return builder.EndObject()")
    lines.append("")
    sha1 = hashlib.sha1()
    sha1.update(name.encode("utf-8"))
    unique_filename = "<FB add_to_builder for %s, %s>" % (
        name,
        sha1.hexdigest(),
    )
    script = "\n".join(lines)
    eval(compile(script, unique_filename, "exec"), globs)

    linecache.cache[unique_filename] = (
        len(script),
        None,
        script.splitlines(True),
        unique_filename,
    )

    return globs["__fb_add_to_builder__"]


def _make_from_bytes_fn(cl) -> Callable:
    """Compile a function to load this model from Flatbuffer bytes."""
    name = cl.__fb_class__.__name__
    globs = {"_fb_cls_loader": getattr(cl.__fb_class__, f"GetRootAs{name}")}
    lines = []
    lines.append("@classmethod")
    lines.append("def __fb_from_bytes__(cls, data):")
    lines.append("    fb_model = _fb_cls_loader(data, 0)")
    lines.append("    return cls.__fb_from_fb__(fb_model)")

    lines.append("")
    sha1 = hashlib.sha1()
    sha1.update(name.encode("utf-8"))
    unique_filename = "<FB from_bytes for %s, %s>" % (name, sha1.hexdigest())
    script = "\n".join(lines)
    eval(compile(script, unique_filename, "exec"), globs)

    linecache.cache[unique_filename] = (
        len(script),
        None,
        script.splitlines(True),
        unique_filename,
    )

    return globs["__fb_from_bytes__"]


def _make_from_fb_fn(
    cl,
    string_fields: List[str],
    optional_strings: List[str],
    byte_fields: List[str],
    optional_bytes: List[str],
    enum_fields: List[str],
    table_fields: List[str],
    optional_tables: List[str],
    lists_of_tables: List[Tuple[str, Type, bool]],
    lists_of_strings: List[Tuple[str, bool]],
    union_fields: List[Tuple[str, List[Type], Type]],
    inlines: List[str],
    lists_of_scalars: List[Tuple[str, Type, Any, bool]],
    lists_of_enums: List[Tuple[str, Type, Any]],
    seqs_of_tables: List[Tuple[str, Type, bool]],
    seqs_of_scalars: List[Tuple[str, Type]],
    seqs_of_enums: List[Tuple[str, Type]],
    seqs_of_strings: List[Tuple[str, bool]],
) -> Callable:
    """Compile a function to init an attrs model from a FB model."""
    name = cl.__fb_class__.__name__
    globs = {}
    lines = []
    list_table_fields = {t[0]: t for t in lists_of_tables}
    seq_table_fields = {t[0]: t for t in seqs_of_tables}
    union_field_names = {t[0]: t for t in union_fields}
    lists_of_scalar_names = {t[0]: t[1] for t in lists_of_scalars}
    seqs_of_scalar_names = {t[0]: t[1] for t in seqs_of_scalars}
    lists_of_enum_names = {t[0]: t for t in lists_of_enums}
    seqs_of_enum_names = {t[0]: t for t in seqs_of_enums}
    lists_of_strings_names = {t[0]: t[1] for t in lists_of_strings}
    seqs_of_strings_names = {t[0]: t[1] for t in seqs_of_strings}

    optionals = {f[0] for f in lists_of_scalars if f[3]}

    from_fb = "__fb_from_fb__"
    lines.append("@classmethod")
    lines.append("def __fb_from_fb__(cls, fb_instance):")
    for fname in optional_bytes:
        norm_field_name = _normalize_fn(fname)
        lines.append(f"    __fb_{fname} = fb_instance.{norm_field_name}AsNumpy()")
    lines.append("    return cls(")
    for field in fields(cl):
        fname = field.name
        norm_field_name = _normalize_fn(fname)
        if fname in string_fields:
            lines.append(f"        fb_instance.{norm_field_name}().decode('utf8'),")
        elif fname in optional_strings:
            lines.append(
                f"        fb_instance.{norm_field_name}().decode('utf8') if fb_instance.{norm_field_name}() is not None else None,"
            )
        elif fname in byte_fields:
            lines.append(f"        fb_instance.{norm_field_name}AsNumpy().tobytes(),")
        elif fname in optional_bytes:
            lines.append(
                f"        __fb_{fname}.tobytes() if not isinstance(__fb_{fname}, int) else None,"
            )
        elif fname in enum_fields:
            enum_name = field.type.__name__
            globs[enum_name] = field.type
            lines.append(f"        {enum_name}(fb_instance.{norm_field_name}()),")
        elif fname in table_fields:
            table_name = field.type.__name__
            globs[table_name] = field.type
            lines.append(
                f"        {table_name}.{from_fb}(fb_instance.{norm_field_name}()),"
            )
        elif fname in optional_tables:
            inner_cl = field.type.__args__[0]
            table_name = inner_cl.__name__
            globs[table_name] = inner_cl
            lines.append(
                f"        {table_name}.{from_fb}(fb_instance.{norm_field_name}()) if fb_instance.{norm_field_name}() is not None else None,"
            )
        elif fname in list_table_fields:
            _, type, is_optional = list_table_fields[fname]
            tn = type.__name__
            globs[tn] = type

            for_ = f"for i in range(fb_instance.{norm_field_name}Length())"
            line = f"        [{tn}.{from_fb}(fb_instance.{norm_field_name}(i)) {for_}]"

            if is_optional:
                # We need the field offset since the Python bindings hide them.
                offset = _get_table_list_offset(cl, fname)
                # We adjust the previous line a little
                line = (
                    f"{line} if int(fb_instance._tab.Offset({offset})) != 0 else None"
                )
            lines.append(f"{line},")
        elif fname in seq_table_fields:
            _, type, is_optional = seq_table_fields[fname]
            tn = type.__name__
            globs[tn] = type
            for_ = f"for i in range(fb_instance.{norm_field_name}Length())"
            line = f"        tuple({tn}.{from_fb}(fb_instance.{norm_field_name}(i)) {for_})"
            if is_optional:
                # We need the field offset since the Python bindings hide them.
                offset = _get_table_list_offset(cl, fname)
                # We adjust the previous line a little
                line = (
                    f"{line} if int(fb_instance._tab.Offset({offset})) != 0 else None"
                )
            lines.append(f"{line},")
        elif fname in lists_of_strings_names:
            is_optional = lists_of_strings_names[fname]
            for_ = f"for i in range(fb_instance.{norm_field_name}Length())"
            line = f"        [fb_instance.{norm_field_name}(i).decode('utf8') {for_}]"
            if is_optional:
                offset = _get_table_list_offset(cl, fname)
                # We adjust the previous line a little
                line = (
                    f"{line} if int(fb_instance._tab.Offset({offset})) != 0 else None"
                )
            lines.append(f"{line},")
        elif fname in seqs_of_strings_names:
            is_optional = seqs_of_strings_names[fname]
            for_ = f"for i in range(fb_instance.{norm_field_name}Length())"
            line = (
                f"        tuple(fb_instance.{norm_field_name}(i).decode('utf8') {for_})"
            )
            if is_optional:
                offset = _get_table_list_offset(cl, fname)
                # We adjust the previous line a little
                line = (
                    f"{line} if int(fb_instance._tab.Offset({offset})) != 0 else None"
                )
            lines.append(f"{line},")
        elif fname in inlines:
            type = field.type
            globs[f"_{fname}_type"] = type
            lines.append(f"        _{fname}_type(fb_instance.{norm_field_name}()),")
        elif fname in union_field_names:
            # We prepare a dictionary to select the proper class at runtime.
            # Then we just grab the proper type and instantiate it.
            _, union_types, union_enum = union_field_names[fname]
            dn = f"_{fname}_union_dict"
            union_resolution_dict = {}
            stripped_union_dict: Dict[str, Any] = {
                k.split("_")[-1]: v for k, v in union_enum.__dict__.items()
            }

            for union_type in [u for u in union_types if u is not none_type]:
                # Unions might be prefixed by a namespace string, depending on how
                # they're defined.
                code = stripped_union_dict[union_type.__name__]
                fb_cls = union_type.__fb_class__
                attr_model = union_type

                def _load_from_content(table, attr_model=attr_model, fb_cls=fb_cls):
                    res = fb_cls()
                    res.Init(table.Bytes, table.Pos)
                    return attr_model.__fb_from_fb__(res)

                union_resolution_dict[code] = _load_from_content

            if none_type in union_types:
                none_code = getattr(union_enum, "NONE")
                union_resolution_dict[none_code] = lambda _: None

            globs[dn] = union_resolution_dict
            lines.append(
                f"        {dn}[fb_instance.{norm_field_name}Type()](fb_instance.{norm_field_name}()),"
            )
        elif fname in lists_of_scalar_names:
            for_ = f"for i in range(fb_instance.{norm_field_name}Length())"
            line = f"        [fb_instance.{norm_field_name}(i) {for_}]"
            if fname in optionals:
                # We need the field offset since the Python bindings hide them.
                _, offset = _get_scalar_list_type(cl, fname)
                # We adjust the previous line a little
                line = (
                    f"{line} if int(fb_instance._tab.Offset({offset})) != 0 else None"
                )
            lines.append(line + ", ")
        elif fname in seqs_of_scalar_names:
            for_ = f"for i in range(fb_instance.{norm_field_name}Length())"
            lines.append(f"        tuple(fb_instance.{norm_field_name}(i) {for_}),")
        elif fname in lists_of_enum_names:
            enum_type = lists_of_enum_names[fname][1]
            dn = f"_{fname}_enum"
            globs[dn] = enum_type
            for_ = f"for i in range(fb_instance.{norm_field_name}Length())"
            lines.append(f"        [{dn}(fb_instance.{norm_field_name}(i)) {for_}],")
        elif fname in seqs_of_enum_names:
            enum_type = seqs_of_enum_names[fname][1]
            dn = f"_{fname}_enum"
            globs[dn] = enum_type
            for_ = f"for i in range(fb_instance.{norm_field_name}Length())"
            lines.append(
                f"        tuple({dn}(fb_instance.{norm_field_name}(i)) {for_}),"
            )
        else:
            raise ValueError(f"Can't handle {fname} (type {field.type}).")

    lines.append("    )")
    lines.append("")

    sha1 = hashlib.sha1()
    sha1.update(name.encode("utf-8"))
    unique_filename = "<FB from_fb for %s, %s>" % (name, sha1.hexdigest())
    script = "\n".join(lines)
    eval(compile(script, unique_filename, "exec"), globs)

    linecache.cache[unique_filename] = (
        len(script),
        None,
        script.splitlines(True),
        unique_filename,
    )

    return globs["__fb_from_fb__"]


def _get_num_slots(fn) -> int:
    class DummyBuilder:
        def StartObject(self, slot_num):
            self.slot_num = slot_num

    d = DummyBuilder()
    fn(d)
    return d.slot_num


def _get_offsets_for_string(fn) -> Tuple[int, int]:
    sentinel = 1024

    class DummyBuilder:
        def PrependUOffsetTRelativeSlot(self, slot_num, offset, default):
            if offset is not sentinel:
                raise ValueError("Failed extracting parameters.")
            self.slot_num = slot_num
            self.default = default

    d = DummyBuilder()
    fn(d, sentinel)
    return d.slot_num, d.default


def _get_scalar_list_type(cl, fname: str) -> Tuple[Any, int]:
    """Fish out the scalar type and field offset for a lists of scalars."""

    class DummyTab:
        def Offset(self, offset):
            self.field_offset = offset
            return 1

        def GetVectorAsNumpy(self, number_type, _):
            self.scalar_type = number_type
            return object()

    tab = DummyTab()
    inst = cl.__fb_class__()
    inst._tab = tab
    norm_field_name = _normalize_fn(fname)
    getattr(inst, f"{norm_field_name}AsNumpy")()

    return tab.scalar_type, tab.field_offset


def _get_table_list_offset(cl, fname: str) -> int:
    class DummyTab:
        def Offset(self, offset):
            self.field_offset = offset
            return 0

    tab = DummyTab()
    inst = cl.__fb_class__()
    inst._tab = tab
    norm_field_name = _normalize_fn(fname)
    getattr(inst, f"{norm_field_name}Length")()

    return tab.field_offset


def _normalize_fn(fname: str) -> str:
    """The name needs to be capitalized.

    If there are two or more consecutive capitals:
    * the cluster is title-cased (isNPC -> isNpc)
    """
    res = sub("[A-Z]+", lambda m: m.group(0).title(), fname)
    # Title-casing needs to happen after the caps processing.
    res = f"{fname[0].upper()}{res[1:]}"
    return res



from platform import python_implementation

from setuptools import find_packages, setup

if python_implementation() != "PyPy":
    from Cython.Build import cythonize

    ext_modules = cythonize("src/flattr/cflattr/*.pyx", annotate=True)
else:
    ext_modules = []

setup(
    name="flattrs",
    version="0.1.16b3",
    install_requires=["attrs", "flatbuffers==23.1.4", "numpy"],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    ext_modules=ext_modules,
    zip_safe=False,
    extras_require={
        "dev": [
            "pytest",
            "hypothesis==6.61.0",
            "cython==0.29.32",
            "coverage",
            "tox==4.0.11",
        ]
    },
    package_data={"": ["*.pxd", "*.pyx", "py.typed"]},
    license="MIT license",
)
